
/* lcc ncr32.c -ltinymain.obj */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
#include <dos.h>


typedef unsigned char byte;
typedef unsigned int  word;
typedef unsigned long dword;

typedef int bool;
#define TRUE  1;
#define FALSE 0;


/* ------------------------------ */
/*                                */
/*            bqb              */
/*                                */
/* ------------------------------ */

/* polynomial */
#define  POLY32 0x04c11db7
#define aPOLY32 0xedb88320

dword CRCTable[256];

void CreateCRCTable()
{
	int i, r;
	dword cx;

	for(i=0; i<256; i++) {
		cx = (dword)i;
		for (r=0; r< 8; r++) {
			if (cx & 1) cx = (cx >> 1) ^ aPOLY32;
			else cx >>=1;
		}
		CRCTable[i] = cx;
	}
}

dword CalculateCRC(int len, byte *buf, dword crc)
{
	crc = ~crc;

	while(len--) crc = CRCTable[(crc ^ *buf++) & 0xff] ^ (crc >> 8);

	return ~crc;
}


/* ------------------------------ */
/*                                */
/*           mdr           */
/*                                */
/* ------------------------------ */

byte buffer[1024*16];
#define BUFFER_SIZE (1024*16)

typedef struct tagNes {
	struct { dword size; char name[24]; } file;
	struct { byte  prg, chr; } size;
	struct { dword prg, chr; } crc;
	byte mapper;
	byte setting;
	byte warning;
	byte compare;
} NES;
#define SETT_MIRROR  1
#define SETT_BACKUP  2
#define SETT_TRAINER 4
#define SETT_4SCREEN 8
#define WARN_HEADER   1
#define WARN_OVERSIZE 2
#define WARN_LESSSIZE 4
#define WARN_DBLPRG   8
#define WARN_DBLCHR   16
#define CMP_INF 1
#define CMP_CHR 2
#define CMP_PRG 4

#define MAXLEN_FILENAME 22

char HeadLine[] = "***- FILESIZE- PROGRAM------- CHARACTER----- INFO------  NAME------------------\n";
char FootLine[] = "---- --------- -------------- -------------- ----------  ----------------------\n";

void OutputInfo(NES *pn, char *s)
{
	if(!s || !pn) return;

	/*
	"***- FILESIZE- PROGRAM------- CHARACTER----- INFO------  NAME------------------\n"
	"%c%c%c   %7ld  %3d: %08lx  %3d: %08lx  %3d(%c%c%c%c)  %c%.22s\n"
	*/

	s += sprintf(s, "%c%c%c   %7ld  ",
		pn->warning & WARN_HEADER    ? 'H' : ' ',
		pn->warning & WARN_OVERSIZE ?
			'O' : (pn->warning & WARN_LESSSIZE ? 'L' : ' '),
		pn->warning & WARN_DBLPRG ? 
			(pn->warning & WARN_DBLCHR ? 'B' : 'P') :
			(pn->warning & WARN_DBLCHR ? 'C' : ' '),
		pn->file.size);
	s += sprintf(s, "%3d: %08lx  %3d: %08lx  %3d(%c%c%c%c) ",
		pn->size.prg, pn->crc.prg,
		pn->size.chr, pn->crc.chr,
		(int)pn->mapper,
		pn->setting & SETT_MIRROR  ? 'V' : 'H',
		pn->setting & SETT_BACKUP  ? 'S' : ' ',
		pn->setting & SETT_TRAINER ? 'T' : ' ',
		pn->setting & SETT_4SCREEN ? '4' : ' ' );
	s += sprintf(s, "%c ",
		pn->compare & CMP_PRG ?
			(pn->compare & CMP_CHR ? (pn->compare & CMP_INF ? 'o' : 'u') : 'x') :
			(pn->compare & CMP_CHR ? 'x' : ' '));
	s += sprintf(s, "%.22s\n",  /* MAXLEN_FILENAME */
		pn->file.name);
}

int InputInfo(NES *pn, char *s)
{
	int i = 0;
	char c[4];

	if(!pn || !s || strlen(s) < 58) return 1;

	pn->warning = 0;
	pn->setting = 0;
	pn->compare = 0;

	if(sscanf(&s[ 5], "%ld", &pn->file.size) != 1) return 1;
	if(sscanf(&s[15], "%d : %lx", &i, &pn->crc.prg) != 2) return 1; else pn->size.prg = (byte)i;
	if(sscanf(&s[30], "%d : %lx", &i, &pn->crc.chr) != 2) return 1; else pn->size.chr = (byte)i;
	if(sscanf(&s[45], "%d(%c%c%c%c)", &i, &c[0], &c[1], &c[2], &c[3]) != 5) return 1;
	else pn->mapper = (byte) i;
	if(c[0] == 'V') pn->setting = 1; else if(c[0] == 'H') pn->setting = 0; else return 1;
	if(c[1] == 'S') pn->setting |= SETT_BACKUP; else if(c[1] != ' ') return 1;
	if(c[2] == 'T') pn->setting |= SETT_TRAINER; else if(c[2] != ' ') return 1;
	if(c[3] == '4') pn->setting |= SETT_4SCREEN; else if(c[3] != ' ') return 1;
	if(sscanf(&s[57], "%22s", pn->file.name) != 1) return 1; /* MAXLEN_FILENAME */

	return 0;
}

byte id_nes[4] = { 0x4e, 0x45, 0x53, 0x1a }; /* NES string */

#define ERR_COULDNOTOPEN 1
#define ERR_NOTNESFILE   2

int GetNESInfo(char *path, NES *ni)
{
	int   ires, i, j;

	int   fh;
	dword fsize;
	dword crc, half1, half2;

	/* ---------------- */
	/* t@CI[v */
	/* ---------------- */
	fh = open(path, O_RDONLY | O_BINARY, S_IREAD | S_IWRITE);
	if(fh == -1) return ERR_COULDNOTOPEN;

	/* -------------- */
	/* wb_ǂݎ */
	/* -------------- */
	lseek(fh, 0, SEEK_SET);
	if(read(fh, buffer, 0x10) != 0x10) return ERR_NOTNESFILE;
	for(i=0; i<4; i++) if(buffer[i] != id_nes[i]) return ERR_NOTNESFILE;

	ni->size.prg = buffer[4];
	ni->size.chr = buffer[5];
	ni->mapper   = (buffer[7] & 0xf0) + (buffer[6]>>4 & 0x0f);
	ni->setting  = buffer[6] & 0x0f;
	ni->warning  = 0;
	ni->compare  = 0;

	/* -------------------- */
	/* wb_̉`FbN */
	/* -------------------- */
	for(i=8; i<16; i++) buffer[8] |= buffer[i];
	if((buffer[7] & 0x0f) || buffer[8]) ni->warning |= WARN_HEADER;

	/* ---------------------------- */
	/* t@CTCY擾ƃ`FbN */
	/* ---------------------------- */
	ni->file.size = lseek(fh, 0, SEEK_END);
	fsize = ni->size.prg * 0x4000 + ni->size.chr * 0x2000 + 0x10;
	if(ni->setting & SETT_TRAINER) fsize += 0x200;
	if(ni->file.size < fsize) {
		ni->warning |= WARN_LESSSIZE;
		ni->size.prg = 0;
		ni->size.chr = 0;
	}
	else if(ni->file.size > fsize) ni->warning |= WARN_OVERSIZE;

	/* ---------- */
	/* bqb擾 */
	/* ---------- */
	fsize = 0x10;
	if(ni->setting & SETT_TRAINER) fsize += 0x200;
	lseek(fh, fsize, SEEK_SET);
	/* program */
	crc = 0;
	j = ni->size.prg>>1;
	for(i=0; i<j; i++) {
		read(fh, buffer, 0x4000);
		crc = CalculateCRC(0x4000, buffer, crc);
	}
	half1 = crc;
	half2 = 0;
	for(; i<ni->size.prg; i++) {
		read(fh, buffer, 0x4000);
		crc = CalculateCRC(0x4000, buffer, crc);
		half2 = CalculateCRC(0x4000, buffer, half2);
	}
	ni->crc.prg = crc;
	if(half1 == half2 && half1 != 0) ni->warning |= WARN_DBLPRG;

	/* character */
	crc = 0;
	j = ni->size.chr>>1;
	for(i=0; i<j; i++) {
		read(fh, buffer, 0x2000);
		crc   = CalculateCRC(0x2000, buffer, crc);
	}
	half1 = crc;
	half2 = 0;
	for(; i<ni->size.chr; i++) {
		read(fh, buffer, 0x2000);
		crc = CalculateCRC(0x2000, buffer, crc);
		half2 = CalculateCRC(0x2000, buffer, half2);
	}
	ni->crc.chr = crc;
	if(half1 == half2 && half1 != 0) ni->warning |= WARN_DBLCHR;

	/* ---------------- */
	/* t@CN[Y */
	/* ---------------- */
	ires = close(fh);

	return 0;
}


/* ------------------------------ */
/*                                */
/*         t@C֘A           */
/*                                */
/* ------------------------------ */

#define MAXNUM_STORE 2048

char far name_store[MAXNUM_STORE * 12];
word     name_index[MAXNUM_STORE * 2];

bool IsKanji1st(byte c)
{
	return (c >= 0x81 && c <= 0x9f) || (c >= 0xe0 && c <= 0xfc);
}

bool GetFilePath(char *s, char *t)
{
	int i, il, ip;

	il = strlen(t);
	ip = -1;

	for(i=0; i<il; i++) {
		if(t[i] == '\\') ip = i;
		else if(IsKanji1st((byte)t[i])) i++;
	}

	if(ip >= 0 && ip < il) {
		strncpy(s, t, ip+1);
		s[ip] = '\0';
		return TRUE;
	} else {
		s[0] = '\0';
		return FALSE;
	}
}

void PutName(int num, char *s)
{
	int ip, il, i;
	char far *ptr;

	if(num < 0 || num >= MAXNUM_STORE) return;

	ptr = name_store + num * 12L;

	ip = 0;
	il = strlen(s);

	for(i=0; i<12; i++) {
		if(*s == '.' && i<8) {
			*ptr = ' ';
		} else if(ip < il) *ptr = *s++, ip++; else *ptr = ' ';
		ptr += 1L;
	}
}

void GetName(int num, char *s)
{
	int i;
	char far *ptr;

	if(num < 0 || num >= MAXNUM_STORE) return;

	ptr = name_store + num * 12L;

	for(i=0; i<12; i++) {
		if(*ptr != ' ') *s++ = *ptr;
		ptr += 1L;
	}
	*s = '\0';
}


int __compare(void *a, void *b)
{
	int i, r;
	char far *x = name_store + (*(word *)a) * 12L;
	char far *y = name_store + (*(word *)b) * 12L;

	for(i=0; i<12; i++) {
		if(r = (int)(*(byte far *)x) - (int)(*(byte far *)y)) return r;
		x += 1L;
		y += 1L;
	}

	return 0;
}

int CreateFileList(char *arg, char *path, int sort)
{
	int num;
	struct find_t tres;

	GetFilePath(path, arg);

	if(_dos_findfirst(arg, _A_NORMAL, &tres) != 0) return 0;
	PutName(0, tres.name);
	name_index[0] = 0;
	num = 1;

	while(_dos_findnext(&tres) == 0) {
		PutName(num, tres.name);
		name_index[num] = num;
		num++;
		if(num >= MAXNUM_STORE) break;
	}

	if(sort) qsort(name_index, (size_t)num, (size_t)2, __compare);

	return num;
}


/* ------------------------------ */
/*                                */
/*          Xg֘A            */
/*                                */
/* ------------------------------ */

#define MAXNUM_LIST 4096
NES far *list_buffer;
int list_num;

int OpenList(char *path)
{
	int   i;
	FILE *fi;
	NES   ne;
	char  sbuf[256];

	i = 0;
	list_num = 0;

	if(!(fi = fopen((const char *)path, "r"))) return 0;

	list_buffer = (NES far *)farmalloc(MAXNUM_LIST * (long)sizeof(NES));
	if(!list_buffer) { fclose(fi); return 0; }

	while(fgets(sbuf, 256, fi)) {
		if(i) {
			if(!strcmp(FootLine, sbuf)) i=0;
			else if(InputInfo(&ne, sbuf)) continue;
			else {
				*(NES far *)(list_buffer + list_num * (long)sizeof(NES)) = ne;
				list_num++;
				if(list_num >= MAXNUM_LIST) break;
			}
		} else if(!strcmp(HeadLine, sbuf)) i = 1;
	}

	fclose(fi);

	return list_num;
}

void CloseList()
{
	farfree(list_buffer);
	list_num = 0;
}

bool CompareList(NES *ni, int replace)
{
	int i, j;
	NES far *ptr;
	byte cmp, prev;

	cmp = 0, prev = 0;
	for(i=0; i<list_num; i++) {
		ptr = list_buffer + i * (long)sizeof(NES);
		cmp = 0;
		if(ni->crc.prg == ptr->crc.prg) cmp |= CMP_PRG;
		if(ni->crc.chr == ptr->crc.chr) cmp |= CMP_CHR;
		if(ni->mapper == ptr->mapper && ni->setting == ptr->setting) cmp |= CMP_INF;
		if(cmp > prev) {
			prev = cmp;
			if(replace) for(j=0; j<24; j++) ni->file.name[j] = ptr->file.name[j];
		}
	}
	ni->compare = prev;
}


/* ------------------------------ */
/*                                */
/*             C             */
/*                                */
/* ------------------------------ */

NES nesi;
char string[256];
char rpath[256];

int main(int argc, char *argv[])
{
	int i, num, nn, mat;
	int tag, lst, rpl, srt;

	printf("\n");
	printf("NCR-32  v1.02 (99/08/04)\n");
	printf("\n");

	tag = lst = rpl = srt = 0;
	if(argc <= 1) {
		printf("ncr32 [/s] [/lXg [/r]] t@C\n");
		printf("\n");
		printf("/s              ʂt@CŃ\[g܂B\n");
		printf("/lXg      r郊Xgt@Cw肵܂B\n");
		printf("/r              XgvAXg̃t@C\܂B\n");
		printf("\n");
		printf("t@C      t@Cw肵܂B\n");
		printf("                ChJ[h *A? gpł܂B\n");
		printf("\n");
		return 0;
	} else {
		for(i=1; i<argc; i++) {
			if((argv[i])[0] == '/'){
				switch((argv[i])[1]) {
				case 's':
				case 'S':
					srt = 1;
					break;
				case 'l':
				case 'L':
					lst = i;
					break;
				case 'r':
				case 'R':
					rpl = 1;
					break;
				}
			} else tag = i;
		}
	}
	if(!tag) return 0;


	if(lst) {
		OpenList(&(argv[lst])[2]);
		mat = 0;
		if(!list_num) { lst = 0; CloseList(); }
	}
	CreateCRCTable();
	num = CreateFileList(argv[tag], rpath, srt);
	nn = 0;

	printf(HeadLine);
	for(i=0; i<num; i++) {
		GetName(name_index[i], nesi.file.name);
		strcpy(string, rpath);
		if(strlen(rpath)) strcat(string, "\\");
		strcat(string, nesi.file.name);
		if(GetNESInfo(string, &nesi)) continue; else nn ++;
		if(lst) {
			CompareList(&nesi, rpl);
			if(nesi.compare & (CMP_CHR | CMP_PRG)) mat++;
		}
		OutputInfo(&nesi, string);
		printf(string);
	}
	printf(FootLine);
	printf("  %4d file(s) found.  %4d file(s) listed.", num, nn);
	if(lst) {
		printf("  %4d file(s) matched.\n", mat);
		CloseList();
	} else printf("\n");
	printf("\n");

	return 0;
}

